def get_values(*names):
    import json
    _all_values = json.loads("""{"num_samp":1,"reset_tipracks":false, "mix_reps2":10, "settling_1":3, "settling_3":3, "settling_drying":8, "wash1_vol":250,"wash2_vol":200, "wash3_vol":100, "elution_vol":50, "binding_buffer_vol":100,"final_vol":80,"heating_module_temp":70,"mag_height_1":3.1,"waste_water_mode":false,"asp_height":0.5,"length_from_side":2,"p300_mount":"right"}""")
    return [_all_values[n] for n in names]

#I added a wash in the original well to pick up any left over beads after RPS wash (incorporated into wash cycle)
#I also added an extra movement of beads from well to temp plate
"""Protocol."""
import os
import csv
from opentrons.types import Point

metadata = {
    'protocolName': 'RNA Extraction with Magnetic Life',
    'author': 'Rami Farawi <rami.farawi@opentrons.com>',
    'source': 'Custom Protocol Request',
    'apiLevel': '2.11'
}


def run(ctx):
    """Protocol."""
    [num_samp, reset_tipracks,  mix_reps2,
     settling_1, settling_3, settling_drying,
     wash1_vol, wash2_vol, wash3_vol, elution_vol,
     binding_buffer_vol, final_vol, heating_module_temp, mag_height_1,
     waste_water_mode, asp_height, length_from_side,
     p300_mount] = get_values(  # noqa: F821
        "num_samp", "reset_tipracks", "mix_reps2",
         "settling_1", "settling_3","settling_drying",
         "wash1_vol", "wash2_vol", "wash3_vol", "elution_vol",
         "binding_buffer_vol", "final_vol",
        "heating_module_temp", "mag_height_1", "waste_water_mode",
        "asp_height", "length_from_side", "p300_mount")

    if not 0 <= num_samp <= 4:
        raise Exception('Please enter a sample number between 1-4')

    # load labware
    mag_mod = ctx.load_module('magnetic module gen2', '7')
    mag_plate = mag_mod.load_labware(
        'innovationlaboratoryproducts_96_wellplate_1000ul', label='Mag Plate')
    reagent_reservoir = ctx.load_labware(
                'nest_12_reservoir_15ml', '1',
                label='Reagent Reservoir')
    waste_reservoir = ctx.load_labware(
                            'nest_1_reservoir_195ml', '4',
                            label='Waste Reservoir')
    tiprack300 = [ctx.load_labware('opentrons_96_tiprack_300ul', slot, '300ul tip rack')
                  for slot in ['2', '3']]
    temp_mod = ctx.load_module('temperature module gen2', '10')
    temp_rack = temp_mod.load_labware(
        'opentrons_96_aluminumblock_biorad_wellplate_200ul',
        label='Temperature tuberack')
    tiprack200 = [ctx.load_labware('opentrons_96_tiprack_300ul', slot,
                                '200Âµl filtertiprack')
               for slot in ['5', '6']]



    # load instrument
    p300 = ctx.load_instrument('p300_multi_gen2',
                               p300_mount,
                               tip_racks=tiprack300)

    tips_300 = [tip_col for tip_col in tiprack300[0].rows()[0]]
    tips_300_park = [tip_col for tip_col in tiprack300[1].rows()[0]]
    regular_tip_count_300 = 0
    park_tip_count_300 = 0

    def pick_up300(park_start_col=0, use_park=False):
        nonlocal regular_tip_count_300
        nonlocal park_tip_count_300

        if regular_tip_count_300 == 11:
            ctx.pause("Replace all 300ul non-filter tip racks")
            regular_tip_count_300 = 0
            p300.pick_up_tip(tips_300[regular_tip_count_300])
            regular_tip_count_300 += 1
        elif use_park:
            p300.pick_up_tip(tips_300_park[park_start_col+park_tip_count_300-1])
            park_tip_count_300 += 1
        else:
            p300.pick_up_tip(tips_300[regular_tip_count_300])
            regular_tip_count_300 += 1

    tips_200 = [tip_col for tip_col in tiprack200[0].rows()[0]]
    tips_200_park = [tip_col for tip_col in tiprack200[1].rows()[0]]
    filter_tip_count = 0
    park_tip_count_200 = 0

    def pick_up_filter(park_start_col=0, use_park=False):
        nonlocal filter_tip_count
        nonlocal park_tip_count_200

        if filter_tip_count == 11:
            ctx.pause("Replace all 200ul filter tip racks")
            filter_tip_count = 0
            p300.pick_up_tip(tips_200[filter_tip_count])
            filter_tip_count += 1
        elif use_park:
            p300.pick_up_tip(tips_200_park[park_start_col+park_tip_count_200-1])
            park_tip_count_200 += 1
        else:
            p300.pick_up_tip(tips_200[filter_tip_count])
            filter_tip_count += 1

    # load reagents
    binding_buffer = reagent_reservoir.wells_by_name()['A1'].bottom(z=2)
    rps_wash_buffer = reagent_reservoir.wells_by_name()['A2'].bottom(z=2)
    rps_wash_buffer2 = reagent_reservoir.wells_by_name()['A3'].bottom(z=2)
    rps_wash_buffer3 = reagent_reservoir.wells_by_name()['A4'].bottom(z=2)
    wash_buffer = reagent_reservoir.wells_by_name()['A5'].bottom(z=2)
    wash_buffer1 = reagent_reservoir.wells_by_name()['A6'].bottom(z=2)
    wash_buffer2 = reagent_reservoir.wells_by_name()['A7'].bottom(z=2)
    wash_buffer3 = reagent_reservoir.wells_by_name()['A8'].bottom(z=2)
    wash_buffer4 = reagent_reservoir.wells_by_name()['A9'].bottom(z=2)
    elution_buffer = reagent_reservoir.wells_by_name()['A11'].bottom(z=2)
    samples = mag_plate.rows()[0][:12:3][:num_samp]
    samples_second_well = mag_plate.rows()[0][1:12:3][:num_samp]
    samples_third_well = mag_plate.rows()[0][2:12:3][:num_samp]
    waste = waste_reservoir.wells_by_name()['A1'].bottom(z=20)
    

    def mixing(vol, index, loc, loc_length_from_side):
        ctx.comment('Mixing')
        side = 0
        mix_loc3 = loc.bottom(z=5).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        p300.mix(10, vol, mix_loc3, rate = 10)

    def mixing_RPS(vol, index, loc, loc_length_from_side):
        ctx.comment('Mixing RPS')
        side = 1
        mix_loc = loc.bottom(z=2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        p300.mix(mix_reps2, vol, mix_loc, rate = 5)

    def mixing_RPS_1(vol, index, loc, loc_length_from_side):
        ctx.comment('Mixing RPS')
        side = 1
        mix_loc_w = loc.bottom(z=3).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        side = 0
        aspirate_loc = loc.bottom(z = 2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        
        for wash in range(15):
            p300.aspirate(vol, aspirate_loc, rate = 5)
            p300.dispense(vol, mix_loc_w, rate = 20)

    def mixing_lysis(vol, index, loc, loc_length_from_side):
        ctx.comment('Mixing RPS')
        side = 1
        mix_loc_w = loc.bottom(z=4).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        side = 0
        aspirate_loc = loc.bottom(z = 2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        
        for wash in range(30):
            p300.aspirate(vol, aspirate_loc, rate = 5)
            p300.dispense(vol, mix_loc_w, rate = 20)

    def aspirate_elution(vol, index, loc, loc_length_from_side):
        ctx.comment('Aspirate Elution')
        side = 0
        aspirate_loc = loc.bottom(z = 3).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        p300.aspirate(vol, aspirate_loc, rate = 5)

    def mixing_RPS_2(vol, index, loc, loc_length_from_side):
        ctx.comment('Mixing RPS')
        side = -1
        mix_loc_w = loc.bottom(z=8).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        side = 0
        aspirate_loc = loc.bottom(z = 2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))

        for wash in range(30):
            p300.aspirate(vol, aspirate_loc, rate = 5)
            p300.dispense(vol, mix_loc_w, rate = 20)
    
    def mixing_elution(vol, index, loc, loc_length_from_side):
        ctx.comment('Mixing Elution')
        side = -1
        mix_loc_w = loc.bottom(z=8).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        side = 0
        aspirate_loc = loc.bottom(z = 2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))

        for wash in range(30):
            p300.aspirate(vol, aspirate_loc, rate = 5)
            p300.dispense(vol, mix_loc_w, rate = 20)
    
    def remove_supernatant(vol, index, loc, loc_length_from_side):
        ctx.comment('Removing Supernatant')

        side = 0
        dispense_loc = loc.bottom(z = 13.5).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        side = -1 
        aspirate_loc = loc.bottom(z = asp_height).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        aspirate_bot = loc.bottom(z = 2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        aspirate_top = loc.bottom(z = 3).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        aspirate_mid = loc.bottom(z = 2.5).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        p300.dispense(0, dispense_loc)
        p300.aspirate(vol, aspirate_top, rate=0.1)
        p300.aspirate(vol, aspirate_mid, rate = 0.1)
        p300.aspirate(vol, aspirate_bot, rate=0.1) 
        p300.dispense(vol*3, waste)

    def remove_supernatant_2(vol, index, loc, loc_length_from_side):
        ctx.comment('Removing Supernatant')

        side = 0
        dispense_loc = loc.bottom(z = 13.5).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        side = 1 
        aspirate_loc = loc.bottom(z = asp_height).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        aspirate_bot = loc.bottom(z = 2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        aspirate_top = loc.bottom(z = 3).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        aspirate_mid = loc.bottom(z = 3.5).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        p300.dispense(0, dispense_loc)
        p300.aspirate(vol, aspirate_top, rate=0.1)
        p300.aspirate(vol, aspirate_mid, rate = 0.1)
        p300.aspirate(vol, aspirate_bot, rate=0.1) 
        p300.dispense(vol*3, waste)

    def dry_samp(vol, index, loc, loc_length_from_side):
        ctx.comment('dry sample')
        side = 0
        dispense_loc = loc.bottom(z = 13.5).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        side = 1
        aspirate_loc = loc.bottom(z = 0.2).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        p300.dispense(0, dispense_loc)
        p300.aspirate(vol, aspirate_loc, rate=0.1)
        p300.dispense(vol, waste)
        p300.drop_tip()

    def dispense_height(vol, index, loc, loc_length_from_side):
        ctx.comment('dispense')
        side = 0
        dispense_loc = loc.bottom(z = 10).move(
                Point(x=(loc.diameter/2-loc_length_from_side)*side))
        p300.dispense(vol, dispense_loc)

    # initialize temp mod and magnetic mod
    temp_mod.set_temperature(heating_module_temp)
    mag_mod.engage(height_from_base=mag_height_1)

    if not waste_water_mode:
        ctx.comment('\n\n\n\nNormal Mode')
        mag_mod.disengage()

   
    # binding
    for wash in range(1):
        pick_up_filter()
        for i, sample in enumerate(samples):
            p300.aspirate(binding_buffer_vol, binding_buffer)
            dispense_height(binding_buffer_vol, i, sample, 2)
        p300.drop_tip()
        for i, sample in enumerate(samples):
            pick_up300()
            mixing(300, i, sample, 1)
            p300.dispense(100, sample)
            p300.touch_tip()
            p300.drop_tip()
        mag_mod.engage(height_from_base=mag_height_1)
        ctx.delay(minutes=5)

        # remove supernatant
        for i, sample in enumerate(samples):
            pick_up300()
            dispense_height(0, i, sample, 2)
            remove_supernatant(100, i, sample, 1)
            remove_supernatant(50, i, sample, 1)
            p300.drop_tip()
        mag_mod.disengage()
        ctx.comment('End Normal Mode\n\n\n\n\n\n\n\n\n')

    if waste_water_mode:
        # remove storage buffer
        ctx.delay(minutes=settling_3)
        for i, sample in enumerate(samples):
            pick_up300(i, use_park=True)
            for _ in range(5):
                p300.aspirate(wash1_vol, sample)
                p300.dispense(wash1_vol, waste)
                p300.blow_out()
            p300.return_tip()
        mag_mod.disengage()
        ctx.comment('\n\n\n')

    #One RPS Wash and move
    ctx.comment('one wash')
    for wash in range(1):
        pick_up300()
        for i, sample in enumerate(samples):
            p300.aspirate(wash1_vol, rps_wash_buffer)
            dispense_height(wash1_vol, i, sample, 2)
        p300.drop_tip()    
        for i, (sample, samp_well2) in enumerate(zip(samples,
                                                 samples_second_well)):
            pick_up300()
            dispense_height(0, i, sample, 2)
            mixing_RPS_1(230, i, sample, 2)
            p300.aspirate(wash1_vol, sample)
            p300.dispense(wash1_vol, samp_well2)
            mixing_RPS_1(230, i, samp_well2, 2)
            p300.touch_tip()
            p300.drop_tip()
        mag_mod.engage(height_from_base=mag_height_1)
        ctx.delay(minutes=3)
        for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
        mag_mod.disengage()

    #One RPS Wash and move
    ctx.comment('one wash')
    for wash in range(1):
        pick_up300()
        for i, sample in enumerate(samples):
            p300.aspirate(wash1_vol, rps_wash_buffer2)
            dispense_height(wash1_vol, i, sample, 2)
        p300.drop_tip()    
        for i, (sample, samp_well2) in enumerate(zip(samples,
                                                 samples_second_well)):
            pick_up300()
            dispense_height(0, i, sample, 2)
            mixing_RPS_1(230, i, sample, 2)
            p300.aspirate(wash1_vol, sample)
            p300.dispense(wash1_vol, samp_well2)
            mixing_RPS_1(230, i, samp_well2, 2)
            p300.touch_tip()
            p300.drop_tip()
        mag_mod.engage(height_from_base=mag_height_1)
        ctx.delay(minutes=3)
        for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
        mag_mod.disengage()

    #One RPS Wash and move
    ctx.comment('one wash')
    for wash in range(1):
        pick_up300()
        for i, sample in enumerate(samples):
            p300.aspirate(wash1_vol, rps_wash_buffer3)
            dispense_height(wash1_vol, i, sample, 2)
        p300.drop_tip()    
        for i, (sample, samp_well2) in enumerate(zip(samples,
                                                 samples_second_well)):
            pick_up300()
            dispense_height(0, i, sample, 2)
            mixing_RPS_1(230, i, sample, 2)
            p300.aspirate(wash1_vol, sample)
            p300.dispense(wash1_vol, samp_well2)
            mixing_RPS_1(230, i, samp_well2, 2)
            p300.touch_tip()
            p300.drop_tip()
        mag_mod.engage(height_from_base=mag_height_1)
        ctx.delay(minutes=3)
        for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
        mag_mod.disengage()

    #One Wash and moving sample to next clean well
    for wash in range(1):
         pick_up300()
         for i, samp_well2 in enumerate(samples_second_well):
             p300.aspirate(wash1_vol, wash_buffer)
             dispense_height(wash1_vol, i, samp_well2, 2)
         p300.drop_tip()
         for i, samp_well2 in enumerate(samples_second_well):
             pick_up300()
             dispense_height(0, i, samp_well2, 2)
             mixing_RPS_1(230, i, samp_well2, 2)
             p300.touch_tip()
             p300.drop_tip()
         mag_mod.engage(height_from_base=mag_height_1)
         ctx.delay(minutes=3)
         for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
         mag_mod.disengage()
        
    #One Wash
    ctx.comment('One wash')
    for wash in range(1):
        pick_up300()
        for i, samp_well2 in enumerate(samples_second_well):
            p300.aspirate(wash1_vol, wash_buffer1)
            dispense_height(wash1_vol, i, samp_well2, 2)
        p300.drop_tip()    
        for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            mixing_RPS_2(230, i, samp_well2, 2)
            p300.touch_tip()
            p300.drop_tip()
        mag_mod.engage(height_from_base=mag_height_1)
        ctx.delay(minutes=3)
        for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
        mag_mod.disengage()

    for wash in range(1):
         pick_up300()
         for i, samp_well2 in enumerate(samples_second_well):
             p300.aspirate(wash1_vol, wash_buffer)
             dispense_height(wash1_vol, i, samp_well2, 2)
         p300.drop_tip()
         for i, samp_well2 in enumerate(samples_second_well):
             pick_up300()
             dispense_height(0, i, samp_well2, 2)
             mixing_RPS_1(230, i, samp_well2, 2)
             p300.touch_tip()
             p300.drop_tip()
         mag_mod.engage(height_from_base=mag_height_1)
         ctx.delay(minutes=3)
         for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
         mag_mod.disengage()

    for wash in range(1):
         pick_up300()
         for i, samp_well2 in enumerate(samples_second_well):
             p300.aspirate(wash1_vol, wash_buffer2)
             dispense_height(wash1_vol, i, samp_well2, 2)
         p300.drop_tip()
         for i, samp_well2 in enumerate(samples_second_well):
             pick_up300()
             dispense_height(0, i, samp_well2, 2)
             mixing_RPS_1(230, i, samp_well2, 2)
             p300.touch_tip()
             p300.drop_tip()
         mag_mod.engage(height_from_base=mag_height_1)
         ctx.delay(minutes=3)
         for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
         mag_mod.disengage()

    for wash in range(1):
         pick_up300()
         for i, samp_well2 in enumerate(samples_second_well):
             p300.aspirate(wash1_vol, wash_buffer3)
             dispense_height(wash1_vol, i, samp_well2, 2)
         p300.drop_tip()
         for i, samp_well2 in enumerate(samples_second_well):
             pick_up300()
             dispense_height(0, i, samp_well2, 2)
             mixing_RPS_1(230, i, samp_well2, 2)
             p300.touch_tip()
             p300.drop_tip()
         mag_mod.engage(height_from_base=mag_height_1)
         ctx.delay(minutes=3)
         for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
         mag_mod.disengage()

    for wash in range(1):
         pick_up300()
         for i, samp_well2 in enumerate(samples_second_well):
             p300.aspirate(wash1_vol, wash_buffer4)
             dispense_height(wash1_vol, i, samp_well2, 2)
         p300.drop_tip()
         for i, samp_well2 in enumerate(samples_second_well):
             pick_up300()
             dispense_height(0, i, samp_well2, 2)
             mixing_RPS_1(230, i, samp_well2, 2)
             p300.touch_tip()
             p300.drop_tip()
         mag_mod.engage(height_from_base=mag_height_1)
         ctx.delay(minutes=3)
         for i, samp_well2 in enumerate(samples_second_well):
            pick_up300()
            dispense_height(0, i, samp_well2, 2)
            remove_supernatant_2(100, i, samp_well2, 1)
            p300.drop_tip()
         mag_mod.disengage()

     #Dispose of Tips
    for i, samp_well2 in enumerate(samples_second_well):
         pick_up300(park_start_col=5, use_park=True)
         p300.drop_tip()
    
     #drying sample
    for i, samp_well2 in enumerate(samples_second_well):
         pick_up_filter()
         dispense_height(0, i, samp_well2, 2)
         dry_samp(wash2_vol, i, samp_well2, 1)
    ctx.delay(minutes=settling_drying)
    mag_mod.disengage()
    
     #Elution buffer addition and movement to hot plate
    for wash in range(1):
         pick_up_filter()
         for samp_well2, source in zip(samples_second_well, temp_rack.rows()[0][6:12:1]):
             p300.aspirate(elution_vol, elution_buffer)
             dispense_height(elution_vol, i, samp_well2, 2)
         p300.drop_tip()
         ctx.delay(minutes=settling_1)
         for samp_well2, source in zip(samples_second_well, temp_rack.rows()[0][6:12:1]):
             pick_up_filter(park_start_col=1, use_park=True)
             dispense_height(0, i, samp_well2, 2)
             mixing_RPS_1(40, i, samp_well2, 1)
             p300.aspirate(80, samp_well2)
             p300.dispense(90, source)
             p300.aspirate(80, samp_well2)
             p300.dispense(90, source)
             p300.blow_out()
             p300.return_tip()
         ctx.delay(minutes=2)
         park_tip_count_200 = 0
    
    
    ctx.comment('\n\n\n')
    
    # elution on Aluminum block
    for i, (source, samp_well3) in enumerate(zip(temp_rack.rows()[0][6:12:1],
                                                  samples_third_well)):
         pick_up_filter(park_start_col=1, use_park=True)
         mixing(80, i, source, 2)
         p300.blow_out()
         p300.return_tip()
    ctx.delay(minutes=3)
    park_tip_count_200 = 0
    for i, (source, samp_well3) in enumerate(zip(temp_rack.rows()[0][6:12:1],
                                                  samples_third_well)):
         pick_up_filter(park_start_col=1, use_park=True)
         mixing(80, i, source, 2)
         p300.aspirate(80, source)
         p300.dispense(80, samp_well3)
         p300.drop_tip()
    ctx.comment('\n\n\n')
    
     # Transfer elution mixture to new clean well on magnetic module
    mag_mod.engage(height_from_base=mag_height_1)
    ctx.delay(minutes=2)
    
    ctx.comment('\n\n\n')